/*	ScrollValBar.c
 *
 *	Control Definition function for a scroll bar for scrolling values,
 *	to be stored as a resource.
 *
 *	Offers these features:
 *		1)	When initialized (through NewControl() or GetNewControl(),
 *			makes contrlRfCon into a ProcPtr which is called everytime
 *			the scroll value is changed (and the indicator redrawn).
 *			For instance, if you are scrolling some value, the routine
 *			pointed to by this ProcPtr could display the value in text
 *			format somewhere.
 *			This (C) routine should take one argument, the ControlHandle,
 *			and should return nothing.
 *			Pass NIL if you do not want any such IndicatorFilterProc.
 *			(GetNewControl() will have to pass NIL, I believe, in the
 *			control resource; to use the filter proc in a dialog,
 *			you have to create a dummy control in the DITL resource,
 *			create this control with NewControl(), and place this
 *			control into the item list.)
 *		2)	When scrolling in the button or page areas, automatically
 *			tracks by scrolling the amount in the LoWord (button) or
 *			HiWord (page) of the contrlRfCon; because of feature (1),
 *			the contrlRfCon needs to be set after the control is
 *			initialized.
 *			This tracking feature, of course, is enabled by passing
 *			-1L as the actionProc argument to TrackControl, with -1L in
 *			the contrlAction field of the control record (see Inside Mac).
 *			(This control definition function initializes the
 *			contrlAction field to -1L; if you want to disable this default
 *			automatic tracking, change the contrlAction field after
 *			initializing the dialog with NewControl() or GetNewControl().)
 *
 *	The following is an example of how to use this CDEF in a dialog:
 *
		#define	DIALOG_RESID		?*	resource id of dialog template	*?
		#define	DUMMY_CTRL_ITEM_NO	?*	number in item list of dummy ctrl item *?
		#define	CDEF_RESID			?*	resource id of this CDEF		*?
	
		DialogPtr		dp;
		int				type;
		Handle			item;
		Rect			box;
		ControlHandle	ch;
		void			IndicatorProc(ControlHandle);
	
		dp =GetNewDialog(DIALOG_RESID, NIL, -1L);
		GetDItem(dp, DUMMY_CTRL_ITEM_NO, &type, &item, &box);
		if (type & ctrlItem) {		?* Checks to make sure control type *?
			DisposeControl((ControlHandle) item);
			ch = NewControl(dp, &box, "\P", Visible, ControlValue, ControlMin,
					ControlMax, CDEF_RESID<<4, (long) &IndicatorProc);
			SetCRefCon(ch, (10L << 16) | (1L) );
	  			?* scroll 10 for page up/down, 1 for button up/down *?
			SetDItem(dp, DUMMY_CTRL_ITEM_NO, ctrlItem+resCtrl,
					(Handle) ch, &box);
		}
 *
 *	By the way, you can change the patterns used (black for arrows/buttons,
 *	dkGray for indicator area) by bit-editing the resource (using ResEdit
 *	or a program like MacTools or MacSnoop, finding the ASCII strings FFFFFFFFFFFFFFFF
 *	(black) and 77DD77DD77DD77DD (dkGray) (very near the end of the
 *	resource) and modifying them.
 *	They are used as arguments to StuffHex (see Inside Mac, QuickDraw).
 *
 *	Created for use in Bouncing Ball program.
 *
 *		-Tyler Abbott, 11/5/90
 */

#include "ScrollValBar.h"

SetSVBIndicatorRects(ControlHandle ch)
{
	pSVBAreas	svba = (pSVBAreas) (**ch).contrlData;
	int			ControlRange = (**ch).contrlMax   - (**ch).contrlMin;

	svba->ThumbPos =	( (((**ch).contrlValue - (**ch).contrlMin)	*
						svba->ThumbExtent) + ControlRange/2 )		/
						ControlRange;
	OffsetRect(&svba->Thumb,
			svba->IndicatorRange.left + svba->ThumbPos - svba->Thumb.left,
			0);
	svba->PageUp.right = svba->Thumb.left;
	svba->PageDown.left = svba->Thumb.right;
}

DrawScrollValBar(ControlHandle ch, int partCode)
{
	pSVBAreas	svba = (pSVBAreas) (**ch).contrlData;

	if ( !(**ch).contrlVis ) return;

	switch(partCode) {
		case DRAW_WHOLE_CONTROL:
							DrawScrollValBar(ch, inUpButton);
							DrawScrollValBar(ch, inDownButton);
							DrawScrollValBar(ch, inThumb);
							break;
		case DRAW_INDICATOR:
		case inThumb:
							SetSVBIndicatorRects(ch);
							if ( (**ch).contrlHilite < 254 ) {
								EraseRect(&svba->Thumb);
								FrameRect(&svba->Thumb);
								DrawScrollValBar(ch, inPageUp);
								DrawScrollValBar(ch, inPageDown);
							}
							else
								FrameRect(&svba->IndicatorRange);
							if (svba->IndicatorFilterProc != 0L)
								(*svba->IndicatorFilterProc)(ch);
							break;
		case inUpButton:
							FillPoly(svba->UpButtonPoly, &svba->ArrowPat);
							FramePoly(svba->UpButtonPoly);
							break;
		case inDownButton:
							FillPoly(svba->DownButtonPoly, &svba->ArrowPat);
							FramePoly(svba->DownButtonPoly);
							break;
		case inPageUp:
							if ( (**ch).contrlHilite < 254 ) {
								FillRect(&svba->PageUp, &svba->IndRangePat);
								FrameRect(&svba->PageUp);
							}
							else
								FrameRect(&svba->IndicatorRange);
							break;
		case inPageDown:
							if ( (**ch).contrlHilite < 254 ) {
								FillRect(&svba->PageDown, &svba->IndRangePat);
								FrameRect(&svba->PageDown);
							}
							else
								FrameRect(&svba->IndicatorRange);
							break;
	}
}

InitializeSVBAreas(ControlHandle ch)
{
	pSVBAreas	svba;
	Rect		r = (**ch).contrlRect;
	int			DiagonalSize = (r.bottom-r.top)/2 - 1;
	int			ExtraSpace = DiagonalSize/3;
	int			Middle;

	svba = (pSVBAreas) NewPtr(sizeof(ScrollValBarAreas) );
	(**ch).contrlData = (Handle) svba;
	SetRect(&svba->UpButton, r.left, r.top,
			r.left+(r.bottom-r.top)+ExtraSpace+1, r.bottom);
	SetRect(&svba->DownButton, r.right-(r.bottom-r.top)-ExtraSpace,
			r.top, r.right, r.bottom);
	SetRect(&svba->IndicatorRange, svba->UpButton.right, r.top,
			svba->DownButton.left, r.bottom);
	svba->PageDown = svba->PageUp = svba->Thumb = svba->IndicatorRange;
	svba->Thumb.right = svba->Thumb.left + (r.bottom-r.top);
	svba->ThumbExtent = (svba->IndicatorRange.right - svba->IndicatorRange.left) -
			(svba->Thumb.right - svba->Thumb.left);
	SetSVBIndicatorRects(ch);

	svba->UpButtonPoly = OpenPoly();
		r = svba->UpButton;
		Middle = (r.top+r.bottom)/2;
		MoveTo(	r.left							,	Middle				);
		Line(	DiagonalSize					,	-DiagonalSize		);
		Line(	0								,	DiagonalSize/2		);
		Line(	 (DiagonalSize+ExtraSpace/2)	,	0					);
		Line(	0								,	DiagonalSize+DiagonalSize%2		);
		Line(	-(DiagonalSize+ExtraSpace/2)	,	0					);
		Line(	0								,	DiagonalSize/2		);
		LineTo(	r.left							,	Middle				);
	ClosePoly();

	svba->DownButtonPoly = OpenPoly();
		r = svba->DownButton;
		Middle = (r.top+r.bottom)/2;
		MoveTo(	r.right							,	Middle				);
		Line(	-DiagonalSize					,	-DiagonalSize		);
		Line(	0								,	DiagonalSize/2		);
		Line(	-(DiagonalSize+ExtraSpace/2)	,	0					);
		Line(	0								,	DiagonalSize+DiagonalSize%2		);
		Line(	 (DiagonalSize+ExtraSpace/2)	,	0					);
		Line(	0								,	DiagonalSize/2		);
		LineTo(	r.right							,	Middle				);
	ClosePoly();

	svba->IndicatorFilterProc = (ProcPtr) GetCRefCon(ch);
	SetCtlAction(ch, (ProcPtr) -1L);

	StuffHex(&svba->ArrowPat, "\PFFFFFFFFFFFFFFFF");
	StuffHex(&svba->IndRangePat, "\P77DD77DD77DD77DD");
}

DisposeSVBAreas(ControlHandle ch)
{
	pSVBAreas	svba = (pSVBAreas) (**ch).contrlData;

	KillPoly(svba->UpButtonPoly);
	KillPoly(svba->DownButtonPoly);
	DisposPtr(svba);
}

int TestSVBControl(ControlHandle ch, long LongPoint)
{
	Point		pt;
	pSVBAreas	svba = (pSVBAreas) (**ch).contrlData;

	if ( (**ch).contrlHilite == 255 )		return(0);

	pt.h = LoWord(LongPoint);
	pt.v = HiWord(LongPoint);
	if ( !PtInRect(pt, &(**ch).contrlRect) )	return(0);

	if ( (**ch).contrlHilite == 254 )		return(254);

	if (PtInRect(pt, &svba->UpButton))		return(inUpButton);
	if (PtInRect(pt, &svba->DownButton))	return(inDownButton);
	if (PtInRect(pt, &svba->Thumb))			return(inThumb);
	if (PtInRect(pt, &svba->PageUp))		return(inPageUp);
	if (PtInRect(pt, &svba->PageDown))		return(inPageDown);

	return(0);
}

CalcSVBRegions(ControlHandle ch, RgnHandle rh)
{
	pSVBAreas	svba;
	Rect		r;

	if ((long) rh & 0x80000000) {
		rh = (RgnHandle) ( (long) rh & 0x00FFFFFF );
		svba = (pSVBAreas) (**ch).contrlData;
		RectRgn(rh, &svba->Thumb);
	}
	else {
		r = (**ch).contrlRect;
		RectRgn(rh, &r);
	}
}

PositionThumb(ControlHandle ch, long param)
{
	pSVBAreas	svba = (pSVBAreas) (**ch).contrlData;

	svba->ThumbPos += LoWord(param);
	(**ch).contrlValue = (**ch).contrlMin +
			( (svba->ThumbPos * ((**ch).contrlMax-(**ch).contrlMin) +
			 svba->ThumbExtent/2) /
			 svba->ThumbExtent );
	DrawScrollValBar(ch, DRAW_INDICATOR);
}

CalcThumbConstraints(ControlHandle ch, long param)
{
	pThumbConstraints	pTC = (pThumbConstraints) param;
	pSVBAreas			svba = (pSVBAreas) (**ch).contrlData;
	Rect				r = svba->IndicatorRange;

	r.left += (pTC->limitRect.left - svba->Thumb.left);
	r.right -= (svba->Thumb.right - pTC->limitRect.left) - 1;
	pTC->limitRect = pTC->slopRect = r;
	InsetRect(&pTC->slopRect, -6, -4);
	pTC->axis = hAxisOnly;
}

int SetRelScrollVal(ControlHandle ch, int dhv)
/*	Sets scroll value, keeping between min and max; returns lines scrolled */
{
	int	theValue, oldValue = GetCtlValue(ch);

	if ( (oldValue + dhv) < GetCtlMin(ch) )
		theValue = GetCtlMin(ch);
	else if ( (oldValue + dhv) > GetCtlMax(ch) )
		theValue = GetCtlMax(ch);
	else
		theValue = oldValue + dhv;
	SetCtlValue(ch, theValue);
	return(theValue-oldValue);
}

SVBAutoTrack(ControlHandle ch, int partCode)
{
	int		ScrollAmount;
	long	ref = GetCRefCon(ch);

	switch (partCode) {
		case inUpButton:	ScrollAmount = -(ref & 0x0000FFFF);	break;
		case inDownButton:	ScrollAmount =  (ref & 0x0000FFFF);	break;
		case inPageUp:		ScrollAmount = -(ref >> 16);		break;
		case inPageDown:	ScrollAmount =  (ref >> 16);	break;
		default:			ScrollAmount =  0;	break;
	}
	SetRelScrollVal(ch,ScrollAmount);
}

pascal long main(int varCode, ControlHandle ch, int msg, long param)
/*	Name this SVB_Control or something if this is not to be stored
 *	as a resource.
 */
{
	switch (msg) {
		case drawCntl:	DrawScrollValBar(ch, (int) param);			break;
		case testCntl:	return( (long) TestSVBControl(ch, param) );
		case calcCRgns:	CalcSVBRegions(ch, (RgnHandle) param);		break;
		case initCntl:	InitializeSVBAreas(ch);						break;
		case dispCntl:	DisposeSVBAreas(ch);						break;
		case posCntl:	PositionThumb(ch, param);					break;
		case thumbCntl:	CalcThumbConstraints(ch, param);			break;
		case dragCntl:	;											break;
		case autoTrack:	SVBAutoTrack(ch, (int) param);				break;
	}
	return(0L);
}
